import std/asyncdispatch, std/unittest, std/os, std/parseopt

import ../src/blobsets

suite "Blob set tests":

  var
    store = newNullStore()
    randomCid = blobHash("")

 # test "zero blob":
 #   check(randomCid == zeroChunk)

  proc randomize() =
    randomCid = blobHash(randomCid.toHex)

  proc testPath(s: BlobSet; root: string): BlobSet =
    for path in walkDirRec(root):
      randomize()
      let
        blob = randomCid
        str = $randomCid
      check(str.toBlobid == randomCid)
      result = waitFor insert(store, s, path, blob, 0)
      var found = false
      apply(store, result, path) do (id: BlobId; size: BiggestInt):
        check(id == randomCid)
        found = true
      check(found)

  test "functional insert":
    let
      a = newBlobSet()
      b = waitFor insert(store, a, "foo", randomCid, 0)
      c = waitFor insert(store, b, "bar", randomCid, 0)
    check(contains(store, b, "foo"))
    check(contains(store, c, "foo"))
    check(contains(store, c, "bar"))
    check(not contains(store, a, "foo"))
    check(not contains(store, a, "bar"))
    check(not contains(store, b, "bar"))

  test "apply":
    var bs = newBlobSet()
    for i in 1..1024:
      let
        name = $i
        blob = blobHash name
      bs = waitFor insert(store, bs, name, blob, 0)
    for i in 1..1024:
      let
        name = $i
        blob = blobHash name
      apply(store, bs, name) do (id: BlobId; size: BiggestInt):
        check(id == blob)

  test "remove":
    var bs = newBlobSet()
    for i in 1..1024:
      let
        name = $i
        blob = blobHash name
      bs = waitFor insert(store, bs, name, blob, 0)
    for i in 1..1024:
      let name = $i
      bs = waitFor remove(store, bs, name)
      check(not contains(store, bs, name))

  test "sets":
    var s = newBlobSet()
    for kind, key, val in getopt():
      if kind == cmdArgument:
        s = s.testPath(key)
    if s.isEmpty:
      s = s.testPath(".")
